/* NVTV Null TV-I2C access -- Dirk Thierbach <dthierbach@gmx.de>
 *
 * This file is part of nvtv, a tool for tv-output on NVidia cards.
 * 
 * nvtv is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * nvtv is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA
 *
 * $Id: tv_null.c,v 1.1.1.1 2003/01/02 04:49:24 rshoff Exp $
 *
 * Contents:
 *
 * Routines to fake null I2C access if no chip is found.
 *
 */

#include "xf86i2c.h"
#include "xfree.h"
#include "debug.h"
#include "tv_null.h"

/* -------- Null chip -------- */

/* This is just to make sure nothing crashes if no chip can be detected. */

void TVNullCreate (TVEncoderObj *this, TVChip chip, I2CDevPtr dev)
{
  ErrorF ("Null chip: create\n");
  this->chip = chip;
  this->dev = dev;
  this->hwconfig = 0; 
}

void TVNullInitRegs (TVEncoderObj *this, int flags)
{
  ErrorF ("Null chip: Init\n");
}

void TVNullSetRegs (TVEncoderObj *this, TVEncoderRegs *r, TVState state)
{
  ErrorF ("Null chip: SetRegs\n");
}

void TVNullSetState (TVEncoderObj *this, TVEncoderRegs *r, TVState state)
{
  ErrorF ("Null chip: SetState\n");
}

TVConnect TVNullGetConnect (TVEncoderObj *this)
{
  ErrorF ("Null chip: Connect\n");
  return CONNECT_NONE;
}

long TVNullGetStatus (TVEncoderObj *this, int index)
{
  ErrorF ("Null chip: Status\n");
  return 0;
}

TVEncoderObj tvNullTemplate = {
  chip:TV_NO_CHIP, dev: NULL, Create: TVNullCreate,
  InitRegs: TVNullInitRegs, SetRegs: TVNullSetRegs, SetState: TVNullSetState,
  GetConnect: TVNullGetConnect, GetStatus: TVNullGetStatus};

