/* NVTV NV TV-I2C access -- Dirk Thierbach <dthierbach@gmx.de>
 *
 * This file is part of nvtv, a tool for tv-output on NVidia cards.
 * 
 * nvtv is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * nvtv is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA
 *
 * $Id: tv_nv.c,v 1.1.1.1 2003/01/02 04:49:20 rshoff Exp $
 *
 * Contents:
 *
 * Routines to control the TV and the I2C bus on a NVidia card, as well
 * as other CRT/TV mode related register programming.
 *
 */

#include <stdio.h>
#include <string.h>
#include <sys/io.h>
#include "debug.h"
#include "xfree.h"
#include "mmio.h"
#include "bitmask.h"
#include "xf86i2c.h"
#include "tv_null.h"
#include "tv_bt.h"
#include "tv_cx.h"
#include "tv_ch.h"
#include "tv_ph.h"
#include "tv_nv.h"
#include "nv_type.h"

#ifdef XBOX_SUPPORT
#include "xbox.h"
#endif

/* -------- Crt access -------- */ 

/* these should go to driver/nv/nv_setup.c in the XFree driver;
   maybe they should be unified with the direct use of VGA_* elsewhere
   in the driver */

U008 readCrtNv (NVPtr pNv, int head, int reg)
{
#ifndef FAKE_CRTC
  VGA_WR08(pNv->riva.PCIO, CRT_INDEX(head), reg);
  return VGA_RD08(pNv->riva.PCIO, CRT_DATA(head));
#else
  FPRINTF ("%02x? ", reg);
  return 0;
#endif
}

void writeCrtNv (NVPtr pNv, int head, int reg, U008 val)
{
#ifdef FAKE_CRTC
  if (reg == 0x1f) {
#endif
  VGA_WR08(pNv->riva.PCIO, CRT_INDEX(head), reg);
  VGA_WR08(pNv->riva.PCIO, CRT_DATA(head), val);
#ifdef FAKE_CRTC
  }
  FPRINTF ("%02x=%02x ", reg, val);
#endif
}

void orCrtNv (NVPtr pNv, int head, int reg, U008 val)
{
#ifndef FAKE_CRTC
  register U008 tmp;

  VGA_WR08(pNv->riva.PCIO, CRT_INDEX(head), reg);
  tmp = VGA_RD08(pNv->riva.PCIO, CRT_DATA(head));
  VGA_WR08(pNv->riva.PCIO, CRT_DATA(head), tmp | val);
#else
  FPRINTF ("%02x|%02x ", reg, val);
#endif
}

void andCrtNv (NVPtr pNv, int head, int reg, U008 val)
{
#ifndef FAKE_CRTC
  register U008 tmp;

  VGA_WR08(pNv->riva.PCIO, CRT_INDEX(head), reg);
  tmp = VGA_RD08(pNv->riva.PCIO, CRT_DATA(head));
  VGA_WR08(pNv->riva.PCIO, CRT_DATA(head), tmp & val);
#else
  FPRINTF ("%02x&%02x ", reg, val);
#endif
}

void andOrCrtNv (NVPtr pNv, int head, int reg, U008 and, U008 or)
{
#ifndef FAKE_CRTC
  register U008 tmp;

  VGA_WR08(pNv->riva.PCIO, CRT_INDEX(head), reg);
  tmp = VGA_RD08(pNv->riva.PCIO, CRT_DATA(head));
  VGA_WR08(pNv->riva.PCIO, CRT_DATA(head), (tmp & and) | or);
#else
  FPRINTF ("%02x&%02x|%02x ", reg, and, or);
#endif
}

inline void unlockCrtNv (NVPtr pNv, int head)
{
#ifndef FAKE_CRTC
  writeCrtNv (pNv, head, 0x1f, 0x57); /* unlock extended registers */
#endif
}

inline void lockCrtNv (NVPtr pNv, int head)
{
#ifndef FAKE_CRTC
#if 0 /**/
  writeCrtNv (pNv, head, 0x1f, 0x99); /* lock extended registers */
#endif
#endif
}

/* the following routines are from nv_setup.c, but they take
   an NVPtr instead of a vgaHWPtr. */

void NVWriteGr(NVPtr pNv, CARD8 index, CARD8 value)
{
    VGA_WR08(pNv->riva.PVIO, VGA_GRAPH_INDEX, index);
    VGA_WR08(pNv->riva.PVIO, VGA_GRAPH_DATA,  value);
}

CARD8 NVReadGr(NVPtr pNv, CARD8 index)
{
    VGA_WR08(pNv->riva.PVIO, VGA_GRAPH_INDEX, index);
    return (VGA_RD08(pNv->riva.PVIO, VGA_GRAPH_DATA));
}

void NVWriteSeq(NVPtr pNv, CARD8 index, CARD8 value)
{
    VGA_WR08(pNv->riva.PVIO, VGA_SEQ_INDEX, index);
    VGA_WR08(pNv->riva.PVIO, VGA_SEQ_DATA,  value);
}

CARD8 NVReadSeq(NVPtr pNv, CARD8 index)
{
    VGA_WR08(pNv->riva.PVIO, VGA_SEQ_INDEX, index);
    return (VGA_RD08(pNv->riva.PVIO, VGA_SEQ_DATA));
}

void NVWriteAttr(NVPtr pNv, int head, CARD8 index, CARD8 value)
{
    volatile CARD8 tmp;
  
    DPRINTF ("NVWriteAttr\n");

    tmp = MMIO_H_IN8(pNv->riva.PCIO, head, 
		     VGA_IOBASE_COLOR + VGA_IN_STAT_1_OFFSET);
#if 0
    if (pVga->paletteEnabled)
        index &= ~0x20;
    else
#endif
        index |= 0x20;
    MMIO_H_OUT8(pNv->riva.PCIO, head, VGA_ATTR_INDEX,  index);
    MMIO_H_OUT8(pNv->riva.PCIO, head, VGA_ATTR_DATA_W, value);
}

CARD8 NVReadAttr(NVPtr pNv, int head, CARD8 index)
{
    volatile CARD8 tmp;

    tmp = MMIO_H_IN8(pNv->riva.PCIO, head, 
		     VGA_IOBASE_COLOR + VGA_IN_STAT_1_OFFSET);
#if 0
    if (pVga->paletteEnabled)
        index &= ~0x20;
    else
#endif
        index |= 0x20;
    MMIO_H_OUT8(pNv->riva.PCIO, head, VGA_ATTR_INDEX, index);
    return MMIO_H_IN8(pNv->riva.PCIO, head, VGA_ATTR_DATA_R);
}

void NVWriteMiscOut(NVPtr pNv, CARD8 value)
{
    VGA_WR08(pNv->riva.PVIO, VGA_MISC_OUT_W, value);
}

CARD8 NVReadMiscOut(NVPtr pNv)
{
    return (VGA_RD08(pNv->riva.PVIO, VGA_MISC_OUT_R));
}

void NVDisablePalette (NVPtr pNv, int head)
{
    volatile CARD8 tmp;

    tmp = VGA_RD08(pNv->riva.PCIO + head * HEAD, 
		   VGA_IOBASE_COLOR + VGA_IN_STAT_1_OFFSET);
    VGA_WR08(pNv->riva.PCIO + head * HEAD, VGA_ATTR_INDEX, 0x20);
}

void NVDacDelay (NVPtr pNv, int head)
{
  volatile CARD8 tmp;

  tmp = VGA_RD08(pNv->riva.PCIO + head * HEAD, 
		 VGA_IOBASE_COLOR + VGA_IN_STAT_1_OFFSET);
  tmp = VGA_RD08(pNv->riva.PCIO + head * HEAD, 
		 VGA_IOBASE_COLOR + VGA_IN_STAT_1_OFFSET);
}

void NVWriteDacMask (NVPtr pNv, int head, CARD8 value)
{
    VGA_WR08(pNv->riva.PDIO + head * HEAD, VGA_DAC_MASK, value);
}

CARD8 NVReadDacMask (NVPtr pNv, int head)
{
    return (VGA_RD08(pNv->riva.PDIO + head * HEAD, VGA_DAC_MASK));
}

void NVWriteDacReadAddr (NVPtr pNv, int head, CARD8 value)
{
    VGA_WR08(pNv->riva.PDIO + head * HEAD, VGA_DAC_READ_ADDR, value);
}

void NVWriteDacWriteAddr (NVPtr pNv, int head, CARD8 value)
{
    VGA_WR08(pNv->riva.PDIO + head * HEAD, VGA_DAC_WRITE_ADDR, value);
}

void NVWriteDacData (NVPtr pNv, int head, CARD8 value)
{
    VGA_WR08(pNv->riva.PDIO + head * HEAD, VGA_DAC_DATA, value);
}

CARD8 NVReadDacData (NVPtr pNv, int head)
{
    return (VGA_RD08(pNv->riva.PDIO + head * HEAD, VGA_DAC_DATA));
}


/* -------- I2C Driver Routines -------- */

#ifdef XBOX_SUPPORT

static unsigned char XboxCmdBuffer;
static int XboxCmdIdx=0;
static int XboxAddr;

static Bool NVTvI2CPutByte(I2CDevPtr d, I2CByte data)
{
    int res;

    /*    fprintf(stdout,"I2C: put addr=0x%02x data=0x%02x\n",XboxAddr,data);*/
    switch(XboxCmdIdx)
    {
       case 0:
             XboxCmdBuffer=data;
             XboxCmdIdx++;
             return TRUE;
       case 1:
             res=I2CTransmitCmdData(XboxAddr,XboxCmdBuffer,data,FALSE);
             if(res<0)
                fprintf(stdout,"I2C: transmit error (0x%08x)\n",res);
             XboxCmdBuffer=0;
             XboxCmdIdx=0;
             return res>=0;
       default:
             fprintf(stdout,"I2C: output error !\n");
             XboxCmdBuffer=0;
             XboxCmdIdx=0;
             return FALSE;
    }
    return FALSE;
}

static Bool NVTvI2CGetByte(I2CDevPtr d, I2CByte *data, Bool last)
{
    int res;

    /*    fprintf(stdout,"I2C: get addr=0x%02x\n",XboxAddr);*/

    if(XboxCmdIdx==1)
    {
        res=I2CTransmitCmdGetReturn(XboxAddr,XboxCmdBuffer);
        *data = res>=0 ? res : 0xFF;
        if(res<0)
            fprintf(stdout,"I2C: recv error (0x%08x)\n",res);
/*      else
            fprintf(stdout,"I2C: recv data=0x%02x\n",*data);*/
    }
    else
    {
        fprintf(stdout,"I2C: unexpected data index=%i\n",XboxCmdIdx);
        res=-1;
    }

    XboxCmdIdx=0;
    XboxCmdBuffer=0;
    return res>=0;
}

static void NVTvI2CStop(I2CDevPtr d)
{
/*    fprintf(stdout,"I2C: stop\n");*/
/*    I2CTransmitStop();*/
    XboxCmdBuffer=0;
    XboxCmdIdx=0;
}

static Bool NVTvI2CAddress(I2CDevPtr d, I2CSlaveAddr addr)
{
    XboxAddr=addr&~1;
    return (XboxAddr==0x8A)&&(d->pI2CBus->BusName[2]=='0');
}
#else
static void NVTvI2CGetBits(I2CBusPtr b, int *clock, int *data)
{
  NVPtr pNv = NVPTR(xf86Screens[b->scrnIndex]);
  unsigned char val;

  /* Get the result. */
  VGA_WR08(pNv->riva.PCIO, CRT_INDEX(pNv->TvHead), b->DriverPrivate.val);
  val = VGA_RD08(pNv->riva.PCIO, CRT_DATA(pNv->TvHead));

  *clock = (val & DDC_SCL_READ_MASK) != 0;
  *data  = (val & DDC_SDA_READ_MASK) != 0;

  DEBUG(ErrorF("NVTvI2CGetBits(%p) val=0x%x, returns clock %d, data %d\n",
	       b, val, *clock, *data));
}

static void NVTvI2CPutBits(I2CBusPtr b, int clock, int data)
{
    NVPtr pNv = NVPTR(xf86Screens[b->scrnIndex]);
    unsigned char val;

    VGA_WR08(pNv->riva.PCIO, CRT_INDEX(pNv->TvHead), b->DriverPrivate.val+1);
    val = VGA_RD08(pNv->riva.PCIO, CRT_DATA(pNv->TvHead)) & 0xf0;
    if (clock)
        val |= DDC_SCL_WRITE_MASK;
    else
        val &= ~DDC_SCL_WRITE_MASK;

    if (data)
        val |= DDC_SDA_WRITE_MASK;
    else
        val &= ~DDC_SDA_WRITE_MASK;

    VGA_WR08(pNv->riva.PCIO, CRT_INDEX(pNv->TvHead), b->DriverPrivate.val+1);
    VGA_WR08(pNv->riva.PCIO, CRT_DATA(pNv->TvHead), val | 0x1);
    
    DEBUG(ErrorF("NVTvI2CPutBits(%p, %d, %d) val=0x%x\n", 
		 b, clock, data, val));
}
#endif

I2CBusPtr NVTvI2CInit (ScrnInfoPtr pScrn, char *name, int private)
{
  I2CBusPtr I2CPtr;

  I2CPtr = xf86CreateI2CBusRec();
  if (I2CPtr) {
    I2CPtr->BusName    = name;
    I2CPtr->scrnIndex  = pScrn->scrnIndex;
#ifdef XBOX_SUPPORT
    I2CPtr->I2CPutByte = NVTvI2CPutByte;
    I2CPtr->I2CGetByte = NVTvI2CGetByte;
    I2CPtr->I2CAddress = NVTvI2CAddress;
    I2CPtr->I2CStop    = NVTvI2CStop;
#else
    I2CPtr->I2CPutBits = NVTvI2CPutBits;
    I2CPtr->I2CGetBits = NVTvI2CGetBits;
#endif
    I2CPtr->AcknTimeout = 5;
    I2CPtr->DriverPrivate.val = private; /* CRTC index for I2C Bus */
  }
  return I2CPtr;
}

Bool NVTvBusInit (ScrnInfoPtr pScrn)
{
  Bool ok;
  NVPtr pNv = NVPTR(pScrn);
  int bus;

#ifdef XBOX_SUPPORT
  iopl(3);
#endif

  pNv->TvChain = NULL;
  pNv->TvHead = 0;
  /* Must be named TV0, TV1, etc.; otherwise I2C_ID won't work */
  pNv->TvBusses[0] = NVTvI2CInit (pScrn, "TV0", 0x3e);
  pNv->TvBusses[1] = NVTvI2CInit (pScrn, "TV1", 0x36);
  pNv->TvBusses[2] = NVTvI2CInit (pScrn, "TV2", 0x50);
  
  ok = TRUE;
  for (bus = 0; bus < NV_NBUS; bus++) {
    if (!xf86I2CBusInit(pNv->TvBusses[bus])) {
      ErrorF ("Cannot init Bus %i\n", bus);
      ok = FALSE;
    }
  }
  return ok;
}

/* -------- CRT -------- */

/* The crt register settings are verified to be the same as the NVidia 
   driver settings, at least for the GeForce2 MX. This includes +1 instead
   of +0 (as in the X 'nv' driver) for horizSync. I do not know if all this 
   works for older/newer cards, since the bitranges might be different.  */

/* FIXME: May use disp_end_skew to horizontal move image? */

void NVSetCrtRegs (NVPtr pNv, int head, TVNvRegs *r)
{
  int horizDisplay    = (r->HDisplay/8)    - 1;
  int horizBlankStart = (r->HBlankStart/8) - 1;
  int horizSyncStart  = (r->HSyncStart/8)  + 1; /* verified */
  int horizSyncEnd    = (r->HSyncEnd/8)    + 1; /* verified */
  int horizBlankEnd   = (r->HBlankEnd/8)   - 1;
  int horizTotal      = (r->HTotal/8)      - 5;
  int vertDisplay     =  r->VDisplay       - 1;
  int vertBlankStart  =  r->VBlankStart    - 1;
  int vertSyncStart   =  r->VSyncStart     - 1;
  int vertSyncEnd     =  r->VSyncEnd       - 1;
  int vertBlankEnd    =  r->VBlankEnd      - 1;
  int vertTotal       =  r->VTotal         - 2;

  unlockCrtNv (pNv, head);
#ifdef FAKE_CRTC
  FPRINTF ("\ncrt[");
#endif

  writeCrtNv (pNv, head, 0x00, Set8Bits(horizTotal) );
  writeCrtNv (pNv, head, 0x01, Set8Bits(horizDisplay));
  writeCrtNv (pNv, head, 0x02, Set8Bits(horizBlankStart));
  writeCrtNv (pNv, head, 0x03, 
	      SetBitField(horizBlankEnd,4:0,4:0) 
	    | SetBit(7)); 
  /* index 03; 6:5 disp_end_skew = 0; 7 vert_retrace_acces = 1 */
  writeCrtNv (pNv, head, 0x04, Set8Bits(horizSyncStart));
  writeCrtNv (pNv, head, 0x05, 
	      SetBitField(horizBlankEnd,5:5,7:7)
	    | SetBitField(horizSyncEnd,4:0,4:0));
  writeCrtNv (pNv, head, 0x06, Set8Bits(vertTotal));
  writeCrtNv (pNv, head, 0x07, 
	      SetBitField(vertTotal,8:8,0:0)
	    | SetBitField(vertDisplay,8:8,1:1)
	    | SetBitField(vertSyncStart,8:8,2:2)
	    | SetBitField(vertDisplay,8:8,3:3)
	    | SetBit(4) /* line_compare_8 */
	    | SetBitField(vertTotal,9:9,5:5)
	    | SetBitField(vertDisplay,9:9,6:6)
	    | SetBitField(vertSyncStart,9:9,7:7));
  writeCrtNv (pNv, head, 0x08, 0x00); 
  writeCrtNv (pNv, head, 0x09, 
	      SetBitField(vertBlankStart,9:9,5:5)
	    | SetBit(6) /* line_compare_9 */
	 /* | ((r->flags & V_DBLSCAN) ? 0x80 : 0x00) */ );
  /* index 09: 4:0 multi_scan (num of lines-1) */
  writeCrtNv (pNv, head, 0x10, Set8Bits(vertSyncStart));
  writeCrtNv (pNv, head, 0x11, 
	      SetBitField(vertSyncEnd,3:0,3:0)
	    | SetBit(5));
  /* index 11: 5 disable_irq2 */
  writeCrtNv (pNv, head, 0x12, Set8Bits(vertDisplay));
  writeCrtNv (pNv, head, 0x14, 0x00); /* underline, double word */
  writeCrtNv (pNv, head, 0x15, Set8Bits(vertBlankStart));
  writeCrtNv (pNv, head, 0x16, Set8Bits(vertBlankEnd));
  /* FIXME */
  /* orCrtNv (pNv, head, 0x17, 0xc3); */ /* mode control */
  writeCrtNv (pNv, head, 0x18, 0xff); /* line_compare_7_0 */
  writeCrtNv (pNv, head, 0x1a, 0x00
                           | (r->HDisplay < 1280) ? 0x04 : 0x00);
  /* NV: 38, 3c / bc */
  /* 1c ?? */
  /* 28 framebuffer format !! */
  /* FIXME TODO: Arbitration 0 (0x1b), 1 (0x20) */
  writeCrtNv (pNv, head, 0x25, 
	      SetBitField(vertTotal,10:10,0:0)
	    | SetBitField(vertDisplay,10:10,1:1)
	    | SetBitField(vertSyncStart,10:10,2:2)
	    | SetBitField(vertBlankStart,10:10,3:3)
	    | SetBitField(horizBlankEnd,6:6,4:4)
	    | (readCrtNv (pNv, head, 0x25) & 0x10)); 
  /* index 25, bit 5: offset_11 */
  writeCrtNv (pNv, head, 0x2D, 
	      SetBitField(horizTotal,8:8,0:0)
	    | SetBitField(horizDisplay,8:8,1:1)
	    | SetBitField(horizBlankStart,8:8,2:2)
	    | SetBitField(horizSyncStart,8:8,3:3)
	    | (readCrtNv (pNv, head, 0x2D) & 0x18)); /* bit 5-4 */

#ifdef FAKE_CRTC
  FPRINTF ("]\n");
#endif
  lockCrtNv (pNv, head);
}


#ifdef XBOX_SUPPORT
void NVSetFPRegs (NVPtr pNv, int head, TVRegs *r)
{
  unlockCrtNv (pNv, head);
#if 0 // original ntsc 640 x 480
  MMIO_H_OUT32(pNv->riva.PRAMDAC, head, 0x800, 479);  // FP_VDISPLAY_END
  MMIO_H_OUT32(pNv->riva.PRAMDAC, head, 0x804, 524);  // FP_VTOTAL
  MMIO_H_OUT32(pNv->riva.PRAMDAC, head, 0x808, 479);  // FP_VCRTC
  MMIO_H_OUT32(pNv->riva.PRAMDAC, head, 0x80c, 490);  // FP_VSYNC_START
  MMIO_H_OUT32(pNv->riva.PRAMDAC, head, 0x810, 493);   // FP_VSYNC_END
  MMIO_H_OUT32(pNv->riva.PRAMDAC, head, 0x814, 0);  // FP_VVALID_START
  MMIO_H_OUT32(pNv->riva.PRAMDAC, head, 0x818, 479);  // FP_VVALID_END
  MMIO_H_OUT32(pNv->riva.PRAMDAC, head, 0x820, 639);  // FP_HDISPLAY_END
  MMIO_H_OUT32(pNv->riva.PRAMDAC, head, 0x824, 775);  // FP_HTOTAL
  MMIO_H_OUT32(pNv->riva.PRAMDAC, head, 0x828, 599);  // FP_HCRTC FIXME
  MMIO_H_OUT32(pNv->riva.PRAMDAC, head, 0x82c, 653);  // FP_HSYNC_START
  MMIO_H_OUT32(pNv->riva.PRAMDAC, head, 0x830, 685);  // FP_HSYNC_END
  MMIO_H_OUT32(pNv->riva.PRAMDAC, head, 0x834, 0);  // FP_HVALID_START
  MMIO_H_OUT32(pNv->riva.PRAMDAC, head, 0x838, 639);  // FP_HVALID_END
#endif

#if 0  // ntsc 640 x 480 Small
  MMIO_H_OUT32(pNv->riva.PRAMDAC, head, 0x800, 479);  // FP_VDISPLAY_END
  MMIO_H_OUT32(pNv->riva.PRAMDAC, head, 0x804, 599);  // FP_VTOTAL
  MMIO_H_OUT32(pNv->riva.PRAMDAC, head, 0x808, 479);  // FP_VCRTC
  MMIO_H_OUT32(pNv->riva.PRAMDAC, head, 0x80c, 526);  // FP_VSYNC_START
  MMIO_H_OUT32(pNv->riva.PRAMDAC, head, 0x810, 529);   // FP_VSYNC_END
  MMIO_H_OUT32(pNv->riva.PRAMDAC, head, 0x814, 0);  // FP_VVALID_START
  MMIO_H_OUT32(pNv->riva.PRAMDAC, head, 0x818, 479);  // FP_VVALID_END
  MMIO_H_OUT32(pNv->riva.PRAMDAC, head, 0x820, 639);  // FP_HDISPLAY_END
  MMIO_H_OUT32(pNv->riva.PRAMDAC, head, 0x824, 783);  // FP_HTOTAL
  MMIO_H_OUT32(pNv->riva.PRAMDAC, head, 0x828, 599);  // FP_HCRTC FIXME
  MMIO_H_OUT32(pNv->riva.PRAMDAC, head, 0x82c, 650);  // FP_HSYNC_START
  MMIO_H_OUT32(pNv->riva.PRAMDAC, head, 0x830, 682);  // FP_HSYNC_END
  MMIO_H_OUT32(pNv->riva.PRAMDAC, head, 0x834, 0);  // FP_HVALID_START
  MMIO_H_OUT32(pNv->riva.PRAMDAC, head, 0x838, 639);  // FP_HVALID_END
#endif

  DPRINTF ("set flat panel\n");

  MMIO_H_OUT32(pNv->riva.PRAMDAC, head, 0x800, r->enc.bt.v_activei-1);  // FP_VDISPLAY_END
  MMIO_H_OUT32(pNv->riva.PRAMDAC, head, 0x804, r->enc.bt.v_linesi-1);  // FP_VTOTAL
  MMIO_H_OUT32(pNv->riva.PRAMDAC, head, 0x808, r->enc.bt.v_activei-1);  // FP_VCRTC
  MMIO_H_OUT32(pNv->riva.PRAMDAC, head, 0x80c, r->enc.bt.v_linesi -1 -r->enc.bt.v_blanki +2);  // FP_VSYNC_START
  MMIO_H_OUT32(pNv->riva.PRAMDAC, head, 0x810, r->enc.bt.v_linesi -1 -r->enc.bt.v_blanki +2 +3);	 // FP_VSYNC_END
  MMIO_H_OUT32(pNv->riva.PRAMDAC, head, 0x814, 0);  // FP_VVALID_START
  MMIO_H_OUT32(pNv->riva.PRAMDAC, head, 0x818, r->enc.bt.v_activei-1);  // FP_VVALID_END
  MMIO_H_OUT32(pNv->riva.PRAMDAC, head, 0x820, r->enc.bt.h_active-1);  // FP_HDISPLAY_END
  MMIO_H_OUT32(pNv->riva.PRAMDAC, head, 0x824, r->enc.bt.h_clki-1);  // FP_HTOTAL
  MMIO_H_OUT32(pNv->riva.PRAMDAC, head, 0x828, r->enc.bt.h_active-41);  // FP_HCRTC FIXME
  MMIO_H_OUT32(pNv->riva.PRAMDAC, head, 0x82c, r->enc.bt.h_clki -1 -r->enc.bt.h_blanki -7);  // FP_HSYNC_START
  MMIO_H_OUT32(pNv->riva.PRAMDAC, head, 0x830, r->enc.bt.h_clki -1 -r->enc.bt.h_blanki -7 +32);  // FP_HSYNC_END
  MMIO_H_OUT32(pNv->riva.PRAMDAC, head, 0x834, 0);  // FP_HVALID_START
  MMIO_H_OUT32(pNv->riva.PRAMDAC, head, 0x838, r->enc.bt.h_active-1);  // FP_HVALID_END
  MMIO_H_OUT32(pNv->riva.PRAMDAC, head, 0x630, 0x2);  // RGB = 0x0,  YCrCb = 0x2
  lockCrtNv (pNv, head);
}
#endif



void NVGetCrtRegs (NVPtr pNv, int head, TVNvRegs *r)
{
  int horizDisplay;
  int horizBlankStart;
  int horizSyncStart;
  int horizSyncEnd;
  int horizBlankEnd;
  int horizTotal;
  int vertDisplay;
  int vertBlankStart;
  int vertSyncStart;
  int vertSyncEnd;
  int vertBlankEnd;
  int vertTotal;

  unlockCrtNv (pNv, head);
  horizTotal      = Set8Bits(readCrtNv(pNv, head, 0x00)) 
                  | SetBitField(readCrtNv(pNv, head, 0x2d),0:0,8:8);
  horizDisplay    = Set8Bits(readCrtNv(pNv, head, 0x01)) 
                  | SetBitField(readCrtNv(pNv, head, 0x2d),1:1,8:8);
  horizBlankStart = Set8Bits(readCrtNv(pNv, head, 0x02))
                  | SetBitField(readCrtNv(pNv, head, 0x2d),2:2,8:8);
  horizBlankEnd   = SetBitField(readCrtNv(pNv, head, 0x03),4:0,4:0)
                  | SetBitField(readCrtNv(pNv, head, 0x05),7:7,5:5)
                  | SetBitField(readCrtNv(pNv, head, 0x25),4:4,6:6);
  horizSyncStart  = Set8Bits(readCrtNv(pNv, head, 0x04)) 
                  | SetBitField(readCrtNv(pNv, head, 0x2d),3:3,8:8);
  horizSyncEnd    = SetBitField(readCrtNv(pNv, head, 0x05),4:0,4:0);

  vertTotal       = Set8Bits(readCrtNv(pNv, head, 0x06)) 
                  | SetBitField(readCrtNv(pNv, head, 0x07),0:0,8:8)
                  | SetBitField(readCrtNv(pNv, head, 0x07),5:5,9:9)
                  | SetBitField(readCrtNv(pNv, head, 0x25),0:0,10:10);
  vertDisplay     = Set8Bits(readCrtNv(pNv, head, 0x12)) 
	          | SetBitField(readCrtNv(pNv, head, 0x07),1:1,8:8)
	          | SetBitField(readCrtNv(pNv, head, 0x07),6:6,9:9)
	          | SetBitField(readCrtNv(pNv, head, 0x25),1:1,10:10);
  vertBlankStart  = Set8Bits(readCrtNv(pNv, head, 0x15))
	          | SetBitField(readCrtNv(pNv, head, 0x07),3:3,8:8)
	          | SetBitField(readCrtNv(pNv, head, 0x09),5:5,9:9)
	          | SetBitField(readCrtNv(pNv, head, 0x25),3:3,10:10);
  vertBlankEnd    = Set8Bits(readCrtNv(pNv, head, 0x16));
  vertSyncStart   = Set8Bits(readCrtNv(pNv, head, 0x10))
	          | SetBitField(readCrtNv(pNv, head, 0x07),2:2,8:8)
	          | SetBitField(readCrtNv(pNv, head, 0x07),7:7,9:9)
	          | SetBitField(readCrtNv(pNv, head, 0x25),2:2,10:10);
  vertSyncEnd     = SetBitField(readCrtNv(pNv, head, 0x11),3:0,3:0);

  /* It is not clear if blank end is relative to blank start or
     total. I have chosen total here, since blank end is equal to
     total normally, anyway. */

  horizBlankEnd  |= SetBitField (horizTotal,8:7,8:7);
  while (horizBlankEnd < horizTotal) horizBlankEnd += SetBit (7);

  horizSyncEnd  |= SetBitField (horizSyncStart,8:5,8:5);
  while (horizSyncEnd < horizSyncStart) horizSyncEnd += SetBit (5);

  vertBlankEnd  |= SetBitField (vertTotal,10:8,10:8);
  while (vertBlankEnd < vertTotal) vertBlankEnd += SetBit (8);

  vertSyncEnd  |= SetBitField (vertSyncStart,10:4,10:4);
  while (vertSyncEnd < vertSyncStart) vertSyncEnd += SetBit (4);

  r->HDisplay    = (horizDisplay    + 1) * 8;   
  r->HBlankStart = (horizBlankStart + 1) * 8;
  r->HBlankEnd   = (horizBlankEnd   + 1) * 8;  
  r->HSyncStart	 = (horizSyncStart  - 1) * 8; 
  r->HSyncEnd  	 = (horizSyncEnd    - 1) * 8;   
  r->HTotal    	 = (horizTotal      + 5) * 8;     
     
  r->VDisplay    = vertDisplay    + 1;
  r->VBlankStart = vertBlankStart + 1;
  r->VBlankEnd   = vertBlankEnd   + 1;
  r->VSyncStart  = vertSyncStart  + 1;
  r->VSyncEnd    = vertSyncEnd    + 1;
  r->VTotal      = vertTotal      + 2;

  r->clock = 0;  
  lockCrtNv (pNv, head);
}

void NVGetTvRegs (NVPtr pNv, int head, TVNvRegs *r)
{
  unlockCrtNv (pNv, head);
  r->latency = SetBitField(readCrtNv(pNv, head, 0x28),5:3,2:0);
  r->slaveVTotal     = MMIO_H_IN32 (pNv->riva.PRAMDAC, head, 0x720);
  r->slaveVSyncStart = MMIO_H_IN32 (pNv->riva.PRAMDAC, head, 0x724);
  r->slaveVSyncEnd   = MMIO_H_IN32 (pNv->riva.PRAMDAC, head, 0x728);
  r->slaveHTotal     = MMIO_H_IN32 (pNv->riva.PRAMDAC, head, 0x72c);
  r->slaveHSyncStart = MMIO_H_IN32 (pNv->riva.PRAMDAC, head, 0x730);
  r->slaveHSyncEnd   = MMIO_H_IN32 (pNv->riva.PRAMDAC, head, 0x734);
  r->slaveUnknown    = MMIO_H_IN32 (pNv->riva.PRAMDAC, head, 0x738);
  /* FIXME: VTotal only bits 10-0, HTotal only bits 10-0/11-0 */
  lockCrtNv (pNv, head);
}

void NVSetTvRegs (NVPtr pNv, int head, TVNvRegs *r)
{

  /* don't set latency */
  MMIO_H_OUT32 (pNv->riva.PRAMDAC, head, 0x720, r->slaveVTotal);
  MMIO_H_OUT32 (pNv->riva.PRAMDAC, head, 0x724, r->slaveVSyncStart);
  MMIO_H_OUT32 (pNv->riva.PRAMDAC, head, 0x728, r->slaveVSyncEnd);
  MMIO_H_OUT32 (pNv->riva.PRAMDAC, head, 0x72c, r->slaveHTotal);
  MMIO_H_OUT32 (pNv->riva.PRAMDAC, head, 0x730, r->slaveHSyncStart);
  MMIO_H_OUT32 (pNv->riva.PRAMDAC, head, 0x734, r->slaveHSyncEnd);
  MMIO_H_OUT32 (pNv->riva.PRAMDAC, head, 0x738, r->slaveUnknown);



}

/* All registers (except actual "switching") */

void NVGetAllRegs (NVPtr pNv, int head, TVRegs *r)
{
  r->devFlags = NVGetDevFlags (pNv, head);
  NVGetCrtRegs (pNv, head, &r->crtc.nv);
  r->crtc.nv.clock = NVGetClock (pNv, head);
  NVGetTvRegs (pNv, head, &r->crtc.nv);
}

void NVSetAllRegs (NVPtr pNv, int head, TVRegs *r)
{
  NVSetCrtRegs (pNv, head, &r->crtc.nv);
  NVModifyClock (pNv, head, r->crtc.nv.clock);
  if (r->devFlags & DEV_TELEVISION)
    NVSetTvRegs (pNv, head, &r->crtc.nv);
#ifdef XBOX_SUPPORT
  NVSetFPRegs(pNv, head, r);
#endif

}

void NVSetCrtLayout (NVPtr pNv, int head, int addr, int ofs)
{
  unlockCrtNv (pNv, head);
    /* crt offset (= crt line increment)? in 0x13,8bit 0x19,10:8,5:7 */

  writeCrtNv (pNv, head, 0x0c, SetBitField(addr,15:8,7:0));
  writeCrtNv (pNv, head, 0x0d, Set8Bits(addr));
  writeCrtNv (pNv, head, 0x13, Set8Bits(ofs));
  writeCrtNv (pNv, head, 0x19, SetBitField(addr,20:16,4:0)
                           | SetBitField(ofs,10:8,7:5));
#if 0 /* FIXME */
  andOrCrtNv (pNv, head, 0x25, ~0xd0); /* VERTICAL_EXTRA */
  andOrCrtNv (pNv, head, 0x2d, ~0xd0); /* HORIZONTAL_EXTRA */
  writeCrtNv (pNv, head, 0x42, ); /* START_ADDR */
#else
  andCrtNv (pNv, head, 0x25, (U008) ~0xd0); /* VERTICAL_EXTRA  */
  andCrtNv (pNv, head, 0x2d, (U008) ~0xd0); /* HORIZONTAL_EXTRA */
  andCrtNv (pNv, head, 0x42, (U008) ~0x0f);
#endif
  lockCrtNv (pNv, head);
}

void NVGetCrtLayout (NVPtr pNv, int head, int *paddr, int *pofs)
{
  int addr, ofs;

  unlockCrtNv (pNv, head);
  addr = Set8Bits(readCrtNv (pNv, head, 0x0d))
       | SetBitField(readCrtNv (pNv, head, 0x0c),7:0,15:8)
       | SetBitField(readCrtNv (pNv, head, 0x19),4:0,20:16)
       | SetBitField(readCrtNv (pNv, head, 0x2d),7:5,25:23)
       | SetBitField(readCrtNv (pNv, head, 0x25),7:6,27:26)
       | SetBitField(readCrtNv (pNv, head, 0x42),3:0,31:28);
  ofs = Set8Bits(readCrtNv (pNv, head, 0x13))
      | SetBitField(readCrtNv (pNv, head, 0x19),7:5,10:8)
      | SetBitField(readCrtNv (pNv, head, 0x25),5:5,11:11);
  lockCrtNv (pNv, head);
  if (paddr) *paddr = addr;
  if (pofs)  *pofs  = ofs;
}

/* -------- Clock / PLL -------- */

#if 0

NV3: 
  chip->MaxVClockFreqKHz = 256000;
NV4, NV10: 
  chip->MaxVClockFreqKHz = 350000;

Get min/max VCO from pNv, init in back_nvidia

(--) NVIDIA(0): Display 0: maximum pixel clock at  8 bpp: 350 MHz
(--) NVIDIA(0): Display 0: maximum pixel clock at 16 bpp: 350 MHz
(--) NVIDIA(0): Display 0: maximum pixel clock at 32 bpp: 300 MHz
(--) NVIDIA(0): Display 1: maximum pixel clock at  8 bpp: 150 MHz
(--) NVIDIA(0): Display 1: maximum pixel clock at 16 bpp: 150 MHz
(--) NVIDIA(0): Display 1: maximum pixel clock at 32 bpp: 150 MHz

static int CalcVClock
(
    int           clockIn,
    int          *clockOut,
    int          *mOut,
    int          *nOut,
    int          *pOut,
    RIVA_HW_INST *chip
)
{
  unsigned lowM, highM, highP;
  unsigned DeltaNew, DeltaOld;
  unsigned VClk, Freq;
  unsigned M, N, P;
    
  DeltaOld = 0xFFFFFFFF;

  VClk     = (unsigned)clockIn;
    
  if (chip->CrystalFreqKHz == 14318)
  {
    lowM  = 8;
    highM = 14 - (chip->Architecture == NV_ARCH_03);
  } else {
    lowM  = 7;
    highM = 13 - (chip->Architecture == NV_ARCH_03);
  }                      

  highP = 4 - (chip->Architecture == NV_ARCH_03);
  for (P = 0; P <= highP; P ++)
  {
    Freq = VClk << P;
    if ((Freq >= 128000) && (Freq <= chip->MaxVClockFreqKHz))
    {
      for (M = lowM; M <= highM; M++)
      {
	N    = (VClk * M / chip->CrystalFreqKHz) << P;
	Freq = (chip->CrystalFreqKHz * N / M) >> P;
	if (Freq > VClk)
	  DeltaNew = Freq - VClk;
	else
	  DeltaNew = VClk - Freq;
	if (DeltaNew < DeltaOld)
	{
	  *mOut     = M;
	  *nOut     = N;
	  *pOut     = P;
	  *clockOut = Freq;
	  DeltaOld  = DeltaNew;
	}
      }
    }
  }
  return (DeltaOld != 0xFFFFFFFF);
}
#endif

long NVCalcMNP (NVArch *arch, long targetFreq, int *mOut, int *nOut, int *pOut)
{
  long bestFreq;
  long bestDelta;
  long delta;
  int minVco, maxVco;
  int maxM;
  int n, m, p;
  long realFreq;

  /* adjust minVco and maxVco */
  minVco = arch->minVco;
  maxVco = arch->maxVco;
  for (n = 0; arch->freqM[n] != 0 && targetFreq >= arch->freqM[n]; n++);
  maxM = arch->maxM[n];
  bestDelta = -1;
  bestFreq = 0;
  for (p = 0; p <= arch->maxP; p++) {
    if ((targetFreq << p) >= minVco && (targetFreq << p) <= maxVco)
    {
      for (m = arch->minM; m <= maxM; m++) {
	n = (((targetFreq << p) * m + 
	      arch->crystalFreq/2) / arch->crystalFreq);
	realFreq = (((arch->crystalFreq * n + m/2) / m) + 
		    ((1 << p) >> 1)) >> p;
	if (realFreq > targetFreq)
	  delta = realFreq - targetFreq;
	else
	  delta = targetFreq - realFreq;
	if (bestDelta == -1 || delta < bestDelta)
	{
	  *mOut	 = m;
	  *nOut	 = n;
	  *pOut	 = p;
	  bestDelta = delta;
	  bestFreq  = realFreq;
	  /* Accept with low relative error if above freqM limit */
	  if (delta == 0 || 
	      (targetFreq >= arch->freqM[0] && (targetFreq / delta > 200)))
	  {
	    return bestFreq;
	  }
	}
      }
    }
  }
  return bestFreq;
}

void NVProgClock (NVPtr pNv, int head, int m, int n, int p)
{
  int reg;
  register CARD32 val;

  DPRINTF ("nv set mnp %i m=%i n=%i p=%i\n", head, m, n, p);
  reg = (head==1) ? 0x520 : 0x508;
  val = MMIO_IN32 (pNv->riva.PRAMDAC, reg);
  val &= ~SetBitField(-1,15:0,15:0);
  val |= SetBitField(m,7:0,7:0) | SetBitField(n,7:0,15:8);
  MMIO_OUT32 (pNv->riva.PRAMDAC, reg, val);
  xf86usleep (1); /* 1 us */
  val &= ~SetBitField(-1,18:16,18:16);
  val |= SetBitField(p,2:0,18:16);
  MMIO_OUT32 (pNv->riva.PRAMDAC, reg, val);
}

long NVGetClock (NVPtr pNv, int head)
{
  int n, m, p;
  register U032 l;
  register long clock;

#ifdef FAKE_MMIO
  return 25000;
#endif
  l = MMIO_IN32 (pNv->riva.PRAMDAC, (head==1) ? 0x520 : 0x508);
  m = SetBitField (l,7:0,7:0);
  n = SetBitField (l,15:8,7:0);
  p = SetBitField (l,18:16,2:0);
  clock = (((pNv->arch.crystalFreq * n + m/2) / m) + ((1 << p) >> 1)) >> p;
  DPRINTF ("nv get clock pll=%08X, m=%i n=%i p=%i, Fref=%li Fclk=%li kHz\n", 
	   l, m, n, p, pNv->arch.crystalFreq, clock);
  return clock;
}

long NVSetClock (NVPtr pNv, int head, long clock)
{
  int m, n, p;

  DPRINTF ("nv set clock %li\n", clock);
  if (clock == 0) return 0;

  clock = NVCalcMNP (&pNv->arch, clock, &m, &n, &p);
  NVProgClock (pNv, head, m, n, p);
  return clock;
}

long NVModifyClock (NVPtr pNv, int head, long clock)
{
  long oldClock, newClock;
  int m, n, p;

  DPRINTF ("nv mod clock %li\n", clock);
  if (clock == 0) return 0;
  oldClock = NVGetClock (pNv, head);
  newClock = NVCalcMNP (&pNv->arch, clock, &m, &n, &p);
  DPRINTF ("nv mod old=%li new=%li\n", oldClock, newClock);
  /* FIXME */
  /* first check against old pll frequency. If better than newly
     calc'd frequency, don't change (may have different limits). */
  NVProgClock (pNv, head, m, n, p);
  return newClock;
}

/* -------- Device enable/disable -------- */

int NVGetDevFlags (NVPtr pNv, int head)
{
  int devFlags;

  unlockCrtNv (pNv, head);
  devFlags = 0;

  if (readCrtNv (pNv, head, 0x28) & 0x80) {
    devFlags |= DEV_TELEVISION;
  } 

  if ((MMIO_H_IN32 (pNv->riva.PRAMDAC, head, 0x608) & 0x00010000) == 0) { 
    devFlags |= DEV_MONITOR;
  }

  /* FIXME: Check for flatpanel properly. Below is only an attempt */
  /* Alternative: DISPEN bit 29-28 in FP_TG_CONTROL 680848 */
  if ((MMIO_H_IN32 (pNv->riva.PRAMDAC, head, 0x880) & 0x30000000) == 0) { 
    devFlags |= DEV_FLATPANEL;
  }

  lockCrtNv (pNv, head);
  return devFlags;
}

void NVEnableMonitor (NVPtr pNv, int head, Bool state)
{
  if (state) {
    DPRINTF ("nv enable monitor\n");

#if 0 /* FIXME experimental */
    /* reset attr ff */
    MMIO_H_IN8 (pNv->riva.PCIO, pNv->TvHead, 0x3da); 
    /* enable disp */
    MMIO_H_OUT8 (pNv->riva.PCIO, pNv->TvHead, 0x3c0, 0x10 | 0x20); 
#endif
#if 1 /* FIXME experimental */
    /* TEST_CONTROL Pwrdown DAC */
    MMIO_H_AND32 (pNv->riva.PRAMDAC, head, 0x608, ~0x00010000); 
#endif
#if 1 /* FIXME standard */
    andCrtNv (pNv, pNv->TvHead, 0x1A, 0x3f);    /* Enable vsync, hsync */
#endif

  } else {

    DPRINTF ("nv disable monitor\n");
#if 1
    /* TEST_CONTROL Pwrdown DAC */
    MMIO_H_OR32 (pNv->riva.PRAMDAC, head, 0x608, 0x00010000); 
#endif
#if 0
    /* reset attr ff */
    MMIO_H_IN8 (pNv->riva.PCIO, pNv->TvHead, 0x3da); 
    /* disable disp */
    MMIO_H_OUT8 (pNv->riva.PCIO, pNv->TvHead, 0x3c0, 0x00);
#endif
#if 0 /* turns both off */
    DPRINTF ("Seq 0x01=%02X\n", NVReadSeq (pNv, 0x01));
    NVWriteSeq (pNv, 0x01, NVReadSeq (pNv, 0x01) | 0x20); /* display off */
#endif
#if 1 /* FIXME standard */
    orCrtNv (pNv, pNv->TvHead, 0x1A, 0x80); /* disable hsync */
#endif
  }
}

/* -------- Device routines -------- */

void NVUpdateTvState (NVPtr pNv);

void NVSetTvDevice (NVPtr pNv, I2CChainPtr chain, Bool init)
{
  DPRINTF ("nv set dev %p %i\n", chain, init);
  TVSetTvEncoder (&pNv->tvEncoder, chain);
  if (init) {
    unlockCrtNv (pNv, pNv->TvHead);
    if (readCrtNv (pNv, pNv->TvHead, 0x28) & 0x80) {
      /* Already in TV mode, don't init */
      tvState = TV_ON;
    } else {
      tvBusOk = TRUE;
      pNv->tvEncoder.InitRegs (&pNv->tvEncoder, PORT_NVIDIA);
    }
    lockCrtNv (pNv, pNv->TvHead);
  }
}

/* 
 * Create chain of tv chips from list of devices on busses
 */

void NVCreateChain (NVPtr pNv)
{
  if (pNv->TvChain) DPRINTF ("NVCreateChain: WARNING! Memory leak\n");
  NVUpdateTvState (pNv);
  pNv->TvChain = TVCreateChain (pNv->TvBusses, NV_NBUS);
}

/*
 * Free chain and device records of all tv chips
 */

void NVDestroyDevices (NVPtr pNv)
{
  if (!pNv) return;
  if (pNv->TvChain) TVDestroyChain (pNv->TvChain);
  pNv->TvChain = NULL;
  TVDestroyDevices (pNv->TvBusses, NV_NBUS);
}

/*
 * Free chain, device records, and all busses
 */

void NVDestroyBusses (NVPtr pNv)
{
  if (!pNv) return;
  NVDestroyDevices (pNv);
  TVDestroyBusses (pNv->TvBusses, NV_NBUS);
}

void NVProbeTvDevices (NVPtr pNv)
{
  if (!pNv) return;
  unlockCrtNv (pNv, pNv->TvHead); 
  TVProbeKnownDevices (pNv->TvBusses, NV_NBUS); 
  /* NVCreateChain also determines chip type and version */
  NVCreateChain (pNv);
#ifdef PROBE_ALL_UNKNOWN
  /* if no dev's found, scan all addresses */
  if (!pNv->TvChain) {
    TVProbeAllDevices (pNv->TvBusses, NV_NBUS);
    NVCreateChain (pNv);
  }
#endif
}

/* -------- Multiple heads -------- */

void NVAssocFeature (NVPtr pNv, int head, int mask)
{
  int h;

  MMIO_H_OR32 (pNv->riva.PCRTC, head, 0x860, mask);
  for (h = 0; h < pNv->arch.heads; h++) {
    if (h != head) {
      MMIO_H_AND32 (pNv->riva.PCRTC, h, 0x860, ~mask);
    }
  }
}

void NVSetVideoHead (NVPtr pNv, int head)
{
  if (pNv->arch.heads == 1) return;
  NVAssocFeature (pNv, head, 0x00001000); /* Video Scaler */
}

void NVSetTvHead (NVPtr pNv, int head)
{
  int h;
  Bool tv;

  if (pNv->arch.heads == 1) return;
  /* FIXME: Don't set if one head is in TV mode */
  /* This won't work for the Flatpanel ... */
  tv = FALSE;
  for (h = 0; h < pNv->arch.heads; h++) {
    unlockCrtNv (pNv, h);
    if (readCrtNv (pNv, h, 0x28) & 0x80) tv = TRUE;
    lockCrtNv (pNv, h);
  }
  if (tv) return;
  NVAssocFeature (pNv, head, 0x00000110); /* TV & DDC */
  tvState = TV_OFF;
  pNv->TvHead = head;
}

void NVInitTvHead (NVPtr pNv)
{
  int h;

  pNv->TvHead = -1;
  for (h = 0; h < pNv->arch.heads; h++) {
    unlockCrtNv (pNv, h);
    if (readCrtNv (pNv, pNv->TvHead, 0x28) & 0x80) pNv->TvHead = h;
    lockCrtNv (pNv, h);
  }
  if (pNv->TvHead < 0) pNv->TvHead = pNv->arch.heads - 1;
  NVAssocFeature (pNv, pNv->TvHead, 0x00000110); /* TV & DDC */
}

void NVCopyHead (NVPtr pNv, int fromHead, int toHead)
{
  DPRINTF ("NVCopyHead %i -> %i %08lX %08lX\n", fromHead, toHead,
	   MMIO_H_IN32 (pNv->riva.PCRTC, fromHead, 0x800), 
	   MMIO_H_IN32 (pNv->riva.PCRTC, fromHead, 0x804));
  /* NV_PCRTC_START */
  MMIO_H_OUT32 (pNv->riva.PCRTC, toHead, 0x800, 
    MMIO_H_IN32 (pNv->riva.PCRTC, fromHead, 0x800)); 
  /* NV_PCRTC_CONFIG */
  MMIO_H_OUT32 (pNv->riva.PCRTC, toHead, 0x804, 
    MMIO_H_IN32 (pNv->riva.PCRTC, fromHead, 0x804));
  MMIO_H_OUT32 (pNv->riva.PRAMDAC, toHead, 0x600, 
    MMIO_H_IN32 (pNv->riva.PRAMDAC, fromHead, 0x600));
  MMIO_H_OUT32 (pNv->riva.PRAMDAC, toHead, 0x848, 
    MMIO_H_IN32 (pNv->riva.PRAMDAC, fromHead, 0x848));
  MMIO_H_OR32 (pNv->riva.PRAMDAC, toHead, 0x880, 0x10000000);
  /* FIXME !! */
  MMIO_H_OUT32 (pNv->riva.PRAMDAC, toHead, 0x814, 
    MMIO_H_IN32 (pNv->riva.PRAMDAC, fromHead, 0x814));
  MMIO_H_OUT32 (pNv->riva.PRAMDAC, toHead, 0x818, 
    MMIO_H_IN32 (pNv->riva.PRAMDAC, fromHead, 0x818));
  unlockCrtNv (pNv, fromHead);
  unlockCrtNv (pNv, toHead);
#if 0
  writeCrtNv (pNv, toHead, 0x44, 0x03); /* FIXME Test */
  writeCrtNv (pNv, toHead, 0x21, 0xfa); /* FIXME Test */
  writeCrtNv (pNv, toHead, 0x26, 0x14); /* FIXME Test */
  writeCrtNv (pNv, toHead, 0x22, readCrtNv (pNv, fromHead, 0x22));
  writeCrtNv (pNv, toHead, 0x23, readCrtNv (pNv, fromHead, 0x23));
#endif
  andOrCrtNv (pNv, toHead, 0x28, ~0x3, 0x3 & readCrtNv (pNv, fromHead, 0x28));
  /* FIXME LCD off */
  writeCrtNv (pNv, toHead, 0x33, 0x00);
  lockCrtNv (pNv, fromHead);
  lockCrtNv (pNv, toHead);
}

void NVCopyGrSeq (NVPtr pNv, int fromHead, int toHead)
{
  DPRINTF ("NVCopyGrSeq %i -> %i\n", fromHead, toHead);
}

void NVInitGrSeq (NVPtr pNv)
{
  NVWriteSeq(pNv, 0x01, 0x01);	/* reenable display */
  NVWriteSeq(pNv, 0x02, 0x0f);	/* if depth != 1 */
  NVWriteSeq(pNv, 0x03, 0x00);	
  NVWriteSeq(pNv, 0x04, 0x0E);	/* if depth < 8 */
  NVWriteSeq(pNv, 0x00, 0x03);	/* End Reset */
  NVWriteGr(pNv, 0x00, 0x00);
  NVWriteGr(pNv, 0x01, 0x00);
  NVWriteGr(pNv, 0x02, 0x00);
  NVWriteGr(pNv, 0x03, 0x00);
  NVWriteGr(pNv, 0x04, 0x00);    /* depth != 1 */
  NVWriteGr(pNv, 0x05, 0x40);    /* depth != 1 && depth != 4 */
  NVWriteGr(pNv, 0x06, 0x05);
  NVWriteGr(pNv, 0x07, 0x0f);
  NVWriteGr(pNv, 0x08, 0xff);
}

void NVInitDac (NVPtr pNv, int head)
{
  int i;

  NVWriteDacMask (pNv, head, 0xFF);
  NVWriteDacWriteAddr (pNv, head, 0x00);
  for (i = 0; i < 256; i++) {
    NVWriteDacData (pNv, head, i);
    NVDacDelay (pNv, head);
    NVWriteDacData (pNv, head, i);
    NVDacDelay (pNv, head);
    NVWriteDacData (pNv, head, i);
    NVDacDelay (pNv, head);
  }
  NVDisablePalette (pNv, head);
}

void NVInitAttr (NVPtr pNv, int head)
{
  int i;

#ifdef XBOX_TIMING_ONLY
  return;
#endif

  /* reset attr ff */
  MMIO_H_IN8 (pNv->riva.PCIO, head, 0x3da); 
  for (i = 0; i <= 0x0f; i++) {
    MMIO_H_OUT8 (pNv->riva.PCIO, head, 0x3c0, i | 0x20);
    MMIO_H_OUT8 (pNv->riva.PCIO, head, 0x3c0, i);
  }
  MMIO_H_OUT8 (pNv->riva.PCIO, head, 0x3c0, 0x10 | 0x20); /* enable disp */
  MMIO_H_OUT8 (pNv->riva.PCIO, head, 0x3c0, 0x41); /* mode control */
  MMIO_H_OUT8 (pNv->riva.PCIO, head, 0x3c0, 0x11 | 0x20); 
  MMIO_H_OUT8 (pNv->riva.PCIO, head, 0x3c0, 0x00); /* overscan value */
  MMIO_H_OUT8 (pNv->riva.PCIO, head, 0x3c0, 0x12 | 0x20); 
  MMIO_H_OUT8 (pNv->riva.PCIO, head, 0x3c0, 0x0f); /* plane enable */
  MMIO_H_OUT8 (pNv->riva.PCIO, head, 0x3c0, 0x13 | 0x20); 
  MMIO_H_OUT8 (pNv->riva.PCIO, head, 0x3c0, 0x00); /* pel panning */
  MMIO_H_OUT8 (pNv->riva.PCIO, head, 0x3c0, 0x14 | 0x20); 
  MMIO_H_OUT8 (pNv->riva.PCIO, head, 0x3c0, 0x00); /* color select */
}

void NVCopyCursor (NVPtr pNv, int fromHead, int toHead)
{
  DPRINTF ("NVCopyCursor %i -> %i\n", fromHead, toHead);
  /* NV_PCRTC_CURSOR */
  MMIO_H_OUT32 (pNv->riva.PCRTC, toHead, 0x80c,
    MMIO_H_IN32 (pNv->riva.PCRTC, fromHead, 0x80c));
  /* NV_PCRTC_CURSOR_CONFIG */
  MMIO_H_OUT32 (pNv->riva.PCRTC, toHead, 0x810,
    MMIO_H_IN32 (pNv->riva.PCRTC, fromHead, 0x810));
  /* CURSOR_CNTRL */
  MMIO_H_OUT32 (pNv->riva.PRAMDAC, toHead, 0x320,
    MMIO_H_IN32 (pNv->riva.PRAMDAC, fromHead, 0x320));
  /* CURSOR_DATA */
  MMIO_H_OUT32 (pNv->riva.PRAMDAC, toHead, 0x324,
    MMIO_H_IN32 (pNv->riva.PRAMDAC, fromHead, 0x324));
  /* CURSOR_DATA */
  MMIO_H_OUT32 (pNv->riva.PRAMDAC, toHead, 0x328,
    MMIO_H_IN32 (pNv->riva.PRAMDAC, fromHead, 0x328));
  unlockCrtNv (pNv, fromHead);
  unlockCrtNv (pNv, toHead);
  writeCrtNv (pNv, toHead, 0x2f, readCrtNv (pNv, fromHead, 0x2f));
  writeCrtNv (pNv, toHead, 0x30, readCrtNv (pNv, fromHead, 0x30));
  writeCrtNv (pNv, toHead, 0x31, readCrtNv (pNv, fromHead, 0x31));
  lockCrtNv (pNv, fromHead);
  lockCrtNv (pNv, toHead);
}

void NVSetCursorPos (NVPtr pNv, int head, int x, int y)
{
  MMIO_H_OUT32 (pNv->riva.PRAMDAC, head, 0x300, 
    SetBitField(x,11:0,11:0) | SetBitField(y,11:0,27:16));
}
  
void NVGetCursorPos (NVPtr pNv, int head, int *x, int *y)
{
  register int v;

  v = MMIO_H_IN32 (pNv->riva.PRAMDAC, head, 0x300);
  *x = SetBitField(v,11:0,11:0);
  *y = SetBitField(v,27:16,11:0);
}

void NVSetVideoPointOut (NVPtr pNv, int x, int y)
{
  MMIO_OUT32 (pNv->riva.PRAMDAC, 0x948,
  MMIO_OUT32 (pNv->riva.PRAMDAC, 0x94C, 
    SetBitField(x,11:0,11:0) | SetBitField(y,11:0,27:16)));
}
  
void NVGetVideoPointOut (NVPtr pNv, int *x, int *y)
{
  register int v;

  v = MMIO_IN32 (pNv->riva.PRAMDAC, 0x948);
  *x = SetBitField(v,11:0,11:0);
  *y = SetBitField(v,27:16,11:0);
}

Bool NVVertIntrEnabled (NVPtr pNv, int head) 
{
  if (MMIO_H_IN32 (pNv->riva.PCRTC, head, 0x140) & SetBit(0)) {
    return TRUE;
  } else {
    return FALSE;
  }
}
  
void NVCopyArb (NVPtr pNv, int fromHead, int toHead)
{
  /* copy aribtration registers */
  unlockCrtNv (pNv, fromHead);
  unlockCrtNv (pNv, toHead);
  writeCrtNv (pNv, toHead, 0x20, readCrtNv (pNv, fromHead, 0x2f));
  writeCrtNv (pNv, toHead, 0x1b, readCrtNv (pNv, fromHead, 0x30));
  lockCrtNv (pNv, fromHead);
  lockCrtNv (pNv, toHead);
}

/* -------- -------- */

/* Intermediate interface to TV. This might not be needed for the
 * XFree driver, and should be moved to a different file in that case.
 */

static int nvCrtIntr;

inline void prepCrtNv (NVPtr pNv, int head)
{
  // unlockCrtNv (pNv, head);
  nvCrtIntr = readCrtNv (pNv, head, 0x11) & 0x10;
  // andCrtNv (pNv, head, 0x11, 0xef);   /* disable vert intr */
  // orCrtNv (pNv, head, 0x21, 0x80);    /* "virtual" changes */
#ifndef XBOX_TIMING_ONLY
  writeCrtNv (pNv, head, 0x21, 0xfa);
#endif
}

inline void commitCrtNv (NVPtr pNv, int head)
{
  // andCrtNv (pNv, head, 0x21, 0x7f);      /* commit changes */
  // orCrtNv (pNv, head, 0x11, nvCrtIntr);  /* enable vert intr again */
}

void NVUpdateTvState (NVPtr pNv)
{
  if (readCrtNv (pNv, pNv->TvHead, 0x28) & 0x80) {
    tvState = TV_ON;
  } else {
    if (tvState != TV_BARS) tvState = TV_OFF;
  }
}

void NVSetTestImage (NVPtr pNv, TVEncoderRegs *r)
{
  /* FIXME chip registers ... */
#ifndef XBOX_TIMING_ONLY
  unlockCrtNv (pNv, pNv->TvHead);
  if (! (readCrtNv (pNv, pNv->TvHead, 0x28) & 0x80)) {
    /* Only do bars if not in tv-mode */
    pNv->tvEncoder.SetState (&pNv->tvEncoder, r, TV_BARS);
  }
  lockCrtNv (pNv, pNv->TvHead);
#endif
}

/* FIXME: Check that if not changing mode, only registers are set,
   i.e. crt/tv regs in tv-mode and crt regs in non-tv mode
*/

void NVSetTvMode (NVPtr pNv, TVRegs *r)
{
  TVState newstate;

  DPRINTF ("NVSetTvMode { ");
  /* FIXME: Don't switch state if not necessary */
  NVUpdateTvState (pNv);
  /* FIXME: Just convert to old 'state' for now */
  newstate = (r->devFlags & DEV_TELEVISION) ? TV_ON : TV_OFF;
  DPRINTF ("nv tv state %i\n", newstate);

  unlockCrtNv (pNv, pNv->TvHead);
  /* FIXME This initialization is always done ... shouldn't be
     necessary for TwinView */
#ifndef XBOX_TIMING_ONLY
  if (pNv->TvHead == 1) {
    DPRINTF ("init 2nd ");
    unlockCrtNv (pNv, 0);
    writeCrtNv (pNv, 0, 0x44, 0x03); 
    NVInitGrSeq (pNv);
    NVInitDac (pNv, pNv->TvHead);
    writeCrtNv (pNv, 0, 0x44, 0x00); 
    NVCopyArb (pNv, 0, pNv->TvHead);
  }
  NVInitAttr (pNv, pNv->TvHead);
#endif

#ifndef XBOX_TIMING_ONLY
#if 1 /* FIXME experimental NV4 blank color */
  MMIO_H_OUT32 (pNv->riva.PRAMDAC, pNv->TvHead, 0x714, 0x00000000);
#endif
#endif

#if 0 /* FIXME experimental */
  /* FIXME Head ... doesn't work yet ... */
  /* Experimental: Wait for vert retrace... */
  MMIO_H_OUT8 (pNv->riva.PCIO, pNv->TvHead, 0x3c4, 0x01);
  MMIO_H_OUT8 (pNv->riva.PCIO, pNv->TvHead, 0x3c5, 0x01);
  /* Wait until in image */
  while (MMIO_H_IN8 (pNv->riva.PCIO, pNv->TvHead, 0x3da) & 0x08);
  /* Wait until retrace starts */
  while (! MMIO_H_IN8 (pNv->riva.PCIO, pNv->TvHead, 0x3da) & 0x08);
#if 0
  /* Wait until again in image */
  while (MMIO_H_IN8 (pNv->riva.PCIO, pNv->TvHead, 0x3da) & 0x08);
#endif
#endif /* FIXME */

  tvBusOk = TRUE;
  if (tvState == TV_ON && newstate != TV_ON) 
  {
    register CARD32 coeff;

#ifndef XBOX_TIMING_ONLY
    prepCrtNv (pNv, pNv->TvHead);
    andCrtNv (pNv, pNv->TvHead, 0x28, 0x7f);
    commitCrtNv (pNv, pNv->TvHead);

    /* power down flatpanel */
#ifndef XBOX_SUPPORT
    MMIO_H_OR32 (pNv->riva.PRAMDAC, pNv->TvHead, 0x880, 0x30000000);
#endif

    coeff = MMIO_IN32 (pNv->riva.PRAMDAC, 0x50C);
    switch (pNv->TvHead)
    {
      case 0:
	coeff &= ~0x00030000; /* VS_PCLK_TV = None */
	break;
      case 1:
	coeff &= ~0x000c0000; /* VS_PCLK_TV2 = None */
	break;
      default:
	/* FIXME error */
        break;
    }
    MMIO_OUT32 (pNv->riva.PRAMDAC, 0x50C, coeff); /* PLL Coeff Sel */
    NVEnableMonitor (pNv, pNv->TvHead, TRUE);

#if 0 /* FIXME experimental */
    {
      int tmp;
      int i;
      U032 a, b;

      /* FIXME Head ... */
      /* Disable slave CRTC mode */
      tmp = readCrtNv (pNv, pNv->TvHead, 0x28) & 0x7b; 
      writeCrtNv (pNv, pNv->TvHead, 0x28, tmp); 
      for (i = 0; i < 25; i++) {
	/* NV_PCRTC_RASTER */
	a = MMIO_H_IN32 (pNv->riva.PCRTC, pNv->TvHead, 0x808); 
	xf86usleep (2000); /* 2ms */
	b = MMIO_H_IN32 (pNv->riva.PCRTC, pNv->TvHead, 0x808); 
	DPRINTF ("off %08x %08x\n", a, b);
	if (a == b) {
	  writeCrtNv (pNv, pNv->TvHead, 0x28, tmp | 0x80); 
	  xf86usleep (1000); /* 2ms */
	  writeCrtNv (pNv, pNv->TvHead, 0x28, tmp); 
	} else {
	  break;
	}
      }
    }
#endif
#endif
  }
  if (newstate != TV_OFF) 
  {
    /* power down flatpanel */
#ifndef XBOX_SUPPORT
    MMIO_H_OR32 (pNv->riva.PRAMDAC, pNv->TvHead, 0x880, 0x30000000);
#endif

#ifndef XBOX_TIMING_ONLY
    /* TV_SETUP: other head *must* be in same mode */ 
    /* This might cause problems with the flatpanel ... */
    MMIO_H_OUT32 (pNv->riva.PRAMDAC, pNv->TvHead ^ 1, 0x700, 0
      | ((r->portHost & PORT_SYNC_DIR) ? 0 : SetBit(0)));

    /* TV_SETUP: this head portHost */ 
    /* FIXME: If changing sync polarity, should do it the right 
       way round, and allow a delay to let the lines become tristate */
    MMIO_H_OUT32 (pNv->riva.PRAMDAC, pNv->TvHead, 0x700, 0 
/* FIXME for scaler separation */
/*
      | (scaler_sep) ? (SetBitField(1,9:8,9:8) | SetBit(12)) : 0)
*/
/* FIXME test this below */ 
/*
      | ((r->portHost & PORT_VSYNC_POLARITY) ? 0 : SetBit(4))
      | ((r->portHost & PORT_HSYNC_POLARITY) ? 0 : SetBit(3))
*/
      | ((r->portHost & PORT_SYNC_DIR) ? 0 : SetBit(0)));
#endif

    pNv->tvEncoder.SetPort (&pNv->tvEncoder, r->portEnc);
    pNv->tvEncoder.SetRegs (&pNv->tvEncoder, &r->enc, newstate);
  }
  if (tvState != newstate) {
    pNv->tvEncoder.SetState (&pNv->tvEncoder, &r->enc, newstate);
  }
  prepCrtNv (pNv, pNv->TvHead);
  NVSetAllRegs (pNv, pNv->TvHead, r);

#ifndef XBOX_TIMING_ONLY
  if (newstate == TV_ON) 
  {
    register CARD32 coeff;

    andCrtNv (pNv, pNv->TvHead, 0x33, ~0x11);
    if (pNv->arch.exact == 0x25) {
      /* FIXME only in slave mode?? */
      writeCrtNv (pNv, pNv->TvHead, 0x53, 0x00); /* FIXME Test this */
      writeCrtNv (pNv, pNv->TvHead, 0x54, 0x00); /* FIXME Test this */
      // MMIO_H_AND32 (pNv->riva.PRAMDAC, pNv-TvHead, 0x8c0, 0xfffffeff);
    }

    /* Check for proper clock frequency if PORT_PCLK_SLAVE */
    if ((r->portHost & PORT_PCLK_MODE) != PORT_PCLK_MASTER) {
      long clock = NVGetClock (pNv, pNv->TvHead);
      if (clock < pNv->tvEncoder.minClock || clock > pNv->tvEncoder.maxClock)
      {
	ERROR ("NVSetTvMode: slave clock out of range\n");
	r->portHost &= ~PORT_PCLK_MODE; /* Master mode */
      }
    }

    /* FIXME TODO: Only one display is driven by the external clock. 
       This will probably not work for flatpanels. */

    coeff = MMIO_IN32 (pNv->riva.PRAMDAC, 0x50C);
    coeff &= ~0x011f0000; /* TVCLK_SRC=Ext, VS_PCLK_TV 1&2, TV_CLK_RATIO=DB1 */
    switch (pNv->TvHead)
    {
      case 0:
	coeff &= ~0x10000000; /* VCLK_RATIO = DB1 */
	if ((r->portHost & PORT_PCLK_MODE) == PORT_PCLK_MASTER) {
	  coeff |=  0x00030000; /* VS_PCLK_TV = BOTH */
	}
	break;
      case 1:
	coeff &= ~0x20000000; /* VCLK2_RATIO = DB1 */
	if ((r->portHost & PORT_PCLK_MODE) == PORT_PCLK_MASTER) {
	  coeff |=  0x000c0000; /* VS_PCLK_TV2 = BOTH */
	}
	break;
      default:
	/* FIXME error */
        break;
    }
    MMIO_OUT32 (pNv->riva.PRAMDAC, 0x50C, coeff); /* PLL Coeff Sel */
    andOrCrtNv (pNv, pNv->TvHead, 0x28, (U008) ~0xf8, 
	      (U008) (0x80 | SetBitField(r->crtc.nv.latency,2:0,5:3)));
    if (! (r->devFlags & DEV_MONITOR)) 
      NVEnableMonitor (pNv, pNv->TvHead, FALSE);
  }
#endif
  commitCrtNv (pNv, pNv->TvHead);
  tvState = newstate;
  lockCrtNv (pNv, pNv->TvHead);
  DPRINTF ("} NVSetTvMode\n");
}

long NVGetTvStatus (NVPtr pNv, int index)
{
  long result; 

  unlockCrtNv (pNv, pNv->TvHead);
  NVUpdateTvState (pNv);
  tvBusOk = TRUE;
  result = pNv->tvEncoder.GetStatus (&pNv->tvEncoder, index);
  lockCrtNv (pNv, pNv->TvHead);
  return result;
}

TVConnect NVGetTvConnect (NVPtr pNv)
{
  TVConnect result;

  unlockCrtNv (pNv, pNv->TvHead);
  NVUpdateTvState (pNv);
  tvBusOk = TRUE;
  result = pNv->tvEncoder.GetConnect (&pNv->tvEncoder);
  lockCrtNv (pNv, pNv->TvHead);
  return result;
}

/* 

      head  TVCLK VCLK VCLK2
<=400  0     DB2  DB1
       1     DB2       DB1
>400   0     DB1  DB1
       1     DB1       DB1


PLL coeff observed:

MX-200 CH: 10000F00 (all res: 1280 -- 640)

Registers used:

PRAMDAC 
  head: 0x600, 0x848, 0x880, 0x814, 0x818, 0x320, 0x324, 0x328, 0x300

*/

/*

CIO:

0x3c0: ARX index data_write
0x3c1: AR data   data_read
0x3c2: INP0
0x3da: INP0_WRITE (MONO/COLOR)  ; w: reset ARX to write
0x3ca: INP0_READ

VIO:
0x102: VSE1
0x3c2: MISC_WR
0x3c3: VSE2
0x3c4: SRX index
0x3c5: SR data
0x3cc: MISC_RD
0x3ce: GRX index
0x3cf: GR data


3c2.w bit 1: MISC enable cpu access to vidmem
3c2.w bit 1: MISC enable/disable display RAM
3c2.w bit 4: MISC disable video

3c3   bit 0: VGA enable/disable
Description: Disables access to display memmory and the other
             VGA's ports

3c4.01 bit 5: display off -- also TV disp off (only show b/w vertical lines)

*/
